extern vec2  iResolution;
extern float iGlobalTime;

extern vec3  CameraPosition;
extern vec3  CameraForward;
extern vec3  CameraUpward;
extern vec3  CameraRight;
extern float CameraAperture;
extern float CameraFocal;

extern vec3  FractalRotation;
extern vec3  FractalOffset;
extern float FractalIterations;

extern vec3  SunVector;
extern float SunSize;

extern bool  Rendering;

#define PI  3.14159
#define TAU 6.28318

//Fractal
const float Scale           = 3.; 
	  mat3  m;//Initialized in main function

//Render
const int   MaxSteps        = 128;  //maximum steps when raymarching
const int   MaxDepth        = 4;    //maximum reflections
const float MaxDistance     = 5.;   //maximum distance from distance field(stops rays from going to nothing)
      float Epsilon         = min(0.001,0.1*pow(Scale,-FractalIterations));//distance below epsilon si counted as a hit
      float BackStep        = Epsilon*3.;//backstep away from surface when reflecting
const float Delta           = 0.00001;//delta used for normal calculation

//Lighting
const vec3  SkyColor        = vec3(.45,.56,1.);//Ambient light
const vec3  SunColor        = vec3(1.);
const float SunSharpness    = 1.;

//Structures
struct Ray{
    vec3 Position;
    vec3 Direction;
};
    
struct SceneInfo{
    float Distance;
    vec3  Color;
};

//Palette Creator
vec3 getPaletteColor( in float t, in vec3 a, in vec3 b, in vec3 c, in vec3 d ){
    return a + b*cos(TAU*(c*t+d) );
}

//Create FractalRotation Matrix
mat3 getFractalRotationMatrix(vec3 xyz){
    vec3 si = sin(xyz);
    vec3 co = cos(xyz);   
    mat3 xrot = mat3(
         1   , 0   , 0   ,
         0   , co.x, si.x,
         0   ,-si.x, co.x
    );       
	mat3 yrot = mat3(
         co.y, 0   ,-si.y,
         0   , 1   , 0   ,
         si.y, 0   , co.y
    );
    mat3 zrot = mat3(
         co.z, si.z, 0   ,
        -si.z, co.z, 0   ,
         0   , 0   , 1
    );      
   return xrot*yrot*zrot;
}

//Menger Sponge Fractal Distance
vec2 MengerSponge(vec3 p){
	float orbit = 1e20;
	float zz = FractalOffset.z * ( Scale - 1.0 ) / Scale;

	#define iteration \
	p = abs ( p );\
	if ( p.x < p.y ) p.xy = p.yx;\
	if ( p.x < p.z ) p.xz = p.zx;\
	if ( p.y < p.z ) p.zy = p.yz;\
	if ( p.z > 0.5 * zz ) p.z = -p.z + zz;\
	p *= m;\
	p *= Scale;\
	p.xy -= FractalOffset.xy * ( Scale - 1.0 );\
	orbit = min ( orbit, length ( p ) );\

	if ( FractalIterations > 0. ){iteration}
	if ( FractalIterations > 1. ){iteration}
	if ( FractalIterations > 2. ){iteration}
	if ( FractalIterations > 3. ){iteration}
	if ( FractalIterations > 4. ){iteration}
	if ( FractalIterations > 5. ){iteration}
	if ( FractalIterations > 6. ){iteration}
	if ( FractalIterations > 7. ){iteration}
	if ( FractalIterations > 8. ){iteration}
	if ( FractalIterations > 9. ){iteration}

	vec3 del = abs ( p ) - FractalOffset;
	float dis = min ( max ( del.x, max ( del.y, del.z ) ), 0.0 ) + length ( max ( del, 0.0 ) );
	      dis *= pow ( Scale, -FractalIterations );

	return vec2 ( dis, orbit ); 
}

//Get distance&material info from scene
SceneInfo getSceneInfo(vec3 pos){
    float Distance     = MaxDistance;
    vec3  Color        = vec3(1);
    
    vec2  FractalInfo = MengerSponge(pos);
    float FractalDistance = FractalInfo.x;
    if(FractalDistance<Distance){
        Distance = FractalDistance;
        Color    = getPaletteColor(sqrt(FractalInfo.y)*2.5, vec3(.5),vec3(.5),vec3(1),vec3(0,.1,.2) );
    }

    float FloorDistance = pos.y+1.;
    if(FloorDistance<Distance){
        Distance = FloorDistance;
        Color    = vec3(.8);
    }
    
    return SceneInfo(
        Distance,
        Color
    );
    
}

//Get normal of scene
vec3 getNormal(vec3 p){
	float d = getSceneInfo(p).Distance;
	return normalize(vec3(
		getSceneInfo(p+vec3(Delta,0,0)).Distance-d,
		getSceneInfo(p+vec3(0,Delta,0)).Distance-d,
		getSceneInfo(p+vec3(0,0,Delta)).Distance-d
	));
}

//Raymarching
int RayMarch(inout Ray iRay){
	for (int i=0;i<MaxSteps;i++){
		float dis = getSceneInfo(iRay.Position).Distance;        
		if(dis<Epsilon && dis>0.) return 1;
        if(dis==MaxDistance) return -1;
		iRay.Position += iRay.Direction*dis;
	}
	return 0;
}

//SkyColor
vec3 getSkyColor(vec3 dir){
    return mix(
        SkyColor,
        SunColor,
        clamp(pow(SunSize/dot(dir-SunVector,dir-SunVector),SunSharpness),0.,1.)
    );  
}

float hash(float seed){
    return fract(sin(seed)*43758.5453123);
}

vec3 getRandomCosWeightedHemiSphereSample(vec3 n, float seed) {
  	vec2  r  = vec2(hash(seed++),hash(seed++));;
    
	vec3  uu = normalize(cross(n,vec3(0,1,1)));
	vec3  vv = cross( uu, n );
	
	float ra = sqrt(r.y);
	float rx = cos(TAU*r.x); 
	float ry = sin(TAU*r.x);
	float rz = sqrt( 1.-r.y );
    
    return normalize( ra*(rx*uu + ry*vv) + rz*n );
}

vec3 getRandomConeSample(vec3 dir, float extent, float seed) {
    vec2  r  =  vec2(hash(seed++),hash(seed++));
    
	vec3  uu = normalize(cross(dir,vec3(0,1,1)));
	vec3  vv = cross(dir, uu);
    
    float rx = cos(TAU*r.x); 
	float ry = sin(TAU*r.x);
    float rz = 1.0-r.y*extent;
    
	float ra = sqrt(1.0-rz*rz);
    
	return ra*(rx*uu+ry*vv)+rz*dir;
}

vec3 Radiance(in Ray iRay, inout float seed){
    vec3 Color  = vec3(1);
    vec3 Direct = vec3(0);
    
    bool noHit = true;
 
    for(int Depth=0;Depth<MaxDepth;Depth++){
        
        int Hit = RayMarch(iRay);
        
        if(Hit==1){//Scene Hit
            
            	noHit = false;
            
                SceneInfo hitInfo   = getSceneInfo(iRay.Position);
                vec3      hitNormal = getNormal(iRay.Position);
                vec3      hitPos    = iRay.Position+hitNormal*BackStep;

				Color *= hitInfo.Color;

                //Direct Sampling
                vec3  SunSampleDir = getRandomConeSample(SunVector,SunSize,seed++);
                float SunLight = dot(hitNormal,SunSampleDir);
                if(SunLight>0.){
                    Ray LightSampleRay = Ray(hitPos,SunSampleDir); 
                    if (RayMarch(LightSampleRay)==-1){
                        Direct += Color*SunLight*SunColor;
                    }
                }
            
                //Random Sampling
                iRay = Ray(
                     hitPos,
                     getRandomCosWeightedHemiSphereSample(hitNormal,seed++)                        
                );
            
        }else if(Hit==-1){//Sky Hit
            
        	return Direct + Color*getSkyColor(iRay.Direction); 
            
        } 
        
    }
    //Ray never hit the sky
    
    if(noHit){//dirty fix for rays that have never hit ANYTHING at all
        return getSkyColor(iRay.Direction);
    }
    
    
    return Direct;
}


vec4 effect( vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords ){
	vec2 fragCoord = vec2(screen_coords.x,iResolution.y-screen_coords.y);
	
    //Fractal Transformation Matrix Initialisation
    m = getFractalRotationMatrix(FractalRotation);
    
	vec3 Color = vec3(0);
	if(Rendering){
		//Seed
		float seed = iGlobalTime + iResolution.y * fragCoord.x / iResolution.x + fragCoord.y / iResolution.y;
		
		//Scaling
		vec2 p = fragCoord + vec2(hash(seed++),hash(seed++))*2.-1.;//anti-aliasing
		vec2 sp  = (p*2.-iResolution.xy)/min(iResolution.x,iResolution.y);

		//Pixel Ray
		vec3 PixelRayDir = normalize(CameraForward+CameraRight*sp.x+CameraUpward*sp.y);
		
		//Focal Point 
		vec3 FocalPoint  = PixelRayDir*CameraFocal;
		
		//Pick random point on Aperture
		float RandomAngle       = TAU*hash(seed++);
		float RandomRadius      = hash(seed++)*CameraAperture;
		vec3  RandomAperturePos = (cos(RandomAngle)*CameraRight+sin(RandomAngle)*CameraUpward)*sqrt(RandomRadius);
		
		//Point on Aperture to Focal Point
		vec3 FinalRayDir = normalize(FocalPoint-RandomAperturePos);
		
		//Final Ray
		Ray PixelRay = Ray(
			CameraPosition+RandomAperturePos,
			FinalRayDir
		);

		//Calculate Radiance
		Color = Radiance(PixelRay,seed);
	}else{		
		//Scaling
		vec2 sp  = (fragCoord*2.-iResolution.xy)/iResolution.y;
		
		//Final Ray
		Ray PixelRay = Ray(
			CameraPosition,
			normalize(CameraForward+CameraRight*sp.x+CameraUpward*sp.y)
		);
		
		//Calculate normal raymarch
		int Hit = RayMarch(PixelRay);
		if(Hit == 1){
			SceneInfo hitInfo   = getSceneInfo(PixelRay.Position);
            vec3      hitNormal = getNormal(PixelRay.Position);

			Color = hitInfo.Color;
			
			float DistanceToHit = distance(PixelRay.Position,CameraPosition);
			float DoFAmount = abs(DistanceToHit-CameraFocal)/CameraFocal*10.;

			Color = mix(vec3(.9,.1,.1),Color,clamp(DoFAmount,0.,1.));
			
			Color *= dot(hitNormal,SunVector)*.5+.5;
			
			Ray ShadowRay = Ray(PixelRay.Position+hitNormal*BackStep,SunVector);
			int ShadowHit = RayMarch(ShadowRay);
			if(ShadowHit==1) Color *= .5;
			
			
		}else{
			Color = getSkyColor(PixelRay.Direction); 
		}
	}
	return vec4(Color,1.);
}